# Adam Rogalewicz
# 
# SL to TA
# distrubuted under GNU GPL licence
#
# This file contains parse_predicate and parse_input functions from the original SLIDE. It is used for the compatibility reasons.

import re
import functions
import join_core
import process_input

def parse_predicate(pred,parsed_preds):
	# parse the input file into the intermediate format using regular expression
	# the return values are 0: single points-to in each rule, 1: empty-rule
	if not (re.search("^[^:=]*::=[^:]*$",pred)):
		raise process_input.InputError("The predicate %s has not a correct form."%pred)
	pred_name=re.sub("\(.*\)::=.*$","",pred)
	if pred_name in parsed_preds:
		raise process_input.InputError("multiple definition of predicate %s"%pred_name)
	pred_par=re.sub('^.*\((.*)\)::=.*$','\\1',pred)
	pred_par=re.split(",",pred_par)
	pred=re.sub("^.*::=","",pred)
	pred=re.split("\|",pred)
	rules=[]
	emp_rules=0
	for rule in pred:
		#------------------------------------------	
		# if the rule is empty (no allocation), parse it separatelly
		if not (re.search("->",rule)):
			# remove "[&]emp" from the rule"
			if re.search("\\*",rule):
				raise process_input.InputError("Parsing %s: now * allowed in empty rules"%rule)
			rule1=re.sub("[\\&]*emp","",rule)
			# add "&" at the beginning for simpler parsing
			rule1=re.sub("^([^\\&])","&\\1",rule1)
			eq_rel=[] # an equality relation between variables
			while(re.search("\\&",rule1)):
				# pick a single equality
				eq=re.sub("^\\&([^\\*\\&]*).*$","\\1",rule1)
				eq=re.split("=",eq)
				rule1=re.sub("^\\&[^\\*\\&]*","",rule1)
				if not len(eq)==2:
					raise process_input.InputError("Parsing %s: only equalities of the form x=y separated by & are allowed"%rule)
				added=0

				for x in eq_rel:
					if (eq[0] in x) and eq[1] in x:
						added=1
						break
					if eq[0] in x:
						x.append(eq[1])
						added=1
						break
					if eq[1] in x:
						x.append(eq[0])
						added=1
						break
				if not added:
					eq_rel.append(eq)
			# remove variables, which are not parameters from eq_rel
			new_eq_rel=[]
			for cl in eq_rel:
				new_cl=[]
				for var in cl:
					if (var in pred_par)or(var=="nil"):
						new_cl.append(var)
				if len(new_cl)>1:
					new_eq_rel.append(new_cl)
			# eq_rel now contains an equality relation between parameters

			rules.append(("",[],[],new_eq_rel))
			emp_rules=1
			continue
			# end of empty rule parsing
		#------------------------------------------	
		# non-empty rule (with points-to predicate)
		if (re.search("^\\\\E",rule)):
			exists=re.sub('^\\\\E([^\.]*)\..*$','\\1',rule)
			exists=re.split(",",exists)
		else:
			exists=[]
		rule=re.sub("\\\\E[^\.]*\.","",rule)
		alloc=re.sub("([^-]*)->.*$","\\1",rule)
		pointsto=re.sub("[^-]*->([^\\*\\&]*).*$","\\1",rule)
		pointsto=re.split(",",pointsto)
		if re.search("->.*\\*.*->",rule):
			raise process_input.InputError("two pointsto in a single predicate rules -- not implemented")
		rule=re.sub("[^-]*->[^\\*\\&]*","",rule)
		equal=[]
		# parse pure part of the rule (i.e. equalities)
		while(re.search("\\&",rule)):
			eq=re.sub("^\\&([^\\*\\&]*).*$","\\1",rule)
			# only disequalities of the form alloc!=nil are allowed. Added just for the SMT-COMP
			if "!=" in eq:
				eq=re.split("!=",eq)
				if (eq[0]==alloc and eq[1]=="nil") or (eq[1]==alloc and eq[0]=="nil"):
					pass
				else:
					raise process_input.InputError("only disequalities of the for alloc!=nil are allowed")
			else:
				eq=re.split("=",eq)
				if not (len(eq)==2):
					raise process_input.InputError("Only equalities between two variables are allowed in pure part")
				if eq[0]==alloc:
					equal.append(eq[1])
				elif eq[1]==alloc:
					equal.append(eq[0])
				else:
					raise process_input.InputError(" only equalities between allocated variable and something are allowed")
			rule=re.sub("^\\&[^\\*\\&]*","",rule)		
		# replace equalities in points-to directly by the allocated node
		for x in range(0,len(pointsto)):
			if pointsto[x] in equal:
				pointsto[x]=alloc

		calles=[]
		while (re.search("\\*",rule)):
			call=re.sub("^\*([^\(]*)\(.*$","\\1",rule)
			call_params=re.sub("^\*[^\(]*\(([^\)]*)\).*$","\\1",rule)
			call_params=re.split(",",call_params)
			# replace equalities in the calls directly by the allocated node
			for x in range(0,len(call_params)):
				if call_params[x] in equal:
					call_params[x]=alloc
			calles.append((call,call_params))
			rule=re.sub("^\*[^\*]*","",rule)

		rules.append((alloc,pointsto,calles,equal))
	parsed_preds[pred_name]=((pred_par,rules))
	return emp_rules
	

#-------------------------------------
# SL2TA
# step 1: load input
# step 2: parse predicates
# step 3: rename_conflicts_with_params (skipped in old version for compatibility reasons)
# step 4: track & eliminate parameters (skipped in old version for compatibility reasons)
# step 5: remove unreachable predicates (skipped in old version for compatibility reasons)
# step 6: sl2ta
#------------------------------------


def parse_input(filename,tiles,globalfreevars):
    #tiles is a global list of actually known tiles
    # it is used to synchronise symbols between multiple automata
    # globalfreevars con
    contains=process_input.load_input(filename)
    preds={}
    if (re.search("^RootCall",contains[0])):
        # NEW version
        # parse the "RootCall"
        # first check, whether there are existentially quentified parameters
        ex_params=[]
        if (re.search("^RootCall\\\\E",contains[0])):
            ex_params=re.sub('^RootCall\\\\E([^\.]*)\..*$','\\1',contains[0])
            ex_params=re.split(",",ex_params)
            contains[0]=re.sub("\\\\E([^\.]*)\.","",contains[0])

        # first check, whether the join operation is needed --- RootCall contains "*"
        rootcall=re.sub('^RootCall',"",contains[0])
        del contains[0]
        top_calls=[]
        pt_seq=0

        while (re.search("\\*",rootcall)):
            # first handle "->" predicate
            if re.search("^[^\*]*->",rootcall):
                lhs=re.sub("^([^-]*)->.*$","\\1",rootcall)
                rhs=re.sub("^[^-]*->([^\*]*)\*.*$","\\1",rootcall)
                rhs=re.split(",",rhs)
                # remove nil and double occurences from rhs
                rhs_not_nil=[]
                for x in rhs:
                    if (not x=="nil") and (not x==lhs) and (not x in rhs_not_nil):
                        rhs_not_nil.append(x)
                # create an unique predicate for the points-to
                pred_name=process_input.get_unique_name("pt%i"%pt_seq,contains)
                pt_seq=pt_seq+1
                top_calls.append((pred_name,[lhs]+rhs_not_nil))
                rule=(lhs,rhs,[],[])
                preds[pred_name]=([lhs]+rhs_not_nil,[rule])
                # do implicit quantification
                process_input.add_implicit_exists(ex_params,globalfreevars,[lhs]+rhs_not_nil)
            else:
                # store the predicate call into top_calls
                call=re.sub("^([^\(]*)\(.*$","\\1",rootcall)
                call_params=re.sub("^[^\(]*\(([^\)]*)\).*$","\\1",rootcall)
                call_params=re.split(",",call_params)
                top_calls.append((call,call_params))
                # do implicit quantification
                process_input.add_implicit_exists(ex_params,globalfreevars,call_params)
            # remove the call
            rootcall=re.sub("^[^\*]*\*","",rootcall)
        if re.search("^[^\*]*->",rootcall):
            lhs=re.sub("^([^-]*)->.*$","\\1",rootcall)
            rhs=re.sub("^[^-]*->([^\*]*)$","\\1",rootcall)
            rhs=re.split(",",rhs)
            # remove nil and double occurences from rhs
            rhs_not_nil=[]
            for x in rhs:
                #if not x=="nil":
                if (not x=="nil") and (not x==lhs) and (not x in rhs_not_nil):
                    rhs_not_nil.append(x)
            # create an unique predicate for the points-to
            pred_name=process_input.get_unique_name("pt%i"%pt_seq,contains)
            pt_seq=pt_seq+1
            top_calls.append((pred_name,[lhs]+rhs_not_nil))
            rule=(lhs,rhs,[],[])
            preds[pred_name]=([lhs]+rhs_not_nil,[rule])
            # do implicit quantification
            process_input.add_implicit_exists(ex_params,globalfreevars,[lhs]+rhs_not_nil)
        else:
            call=re.sub("^([^\(]*)\(.*$","\\1",rootcall)
            call_params=re.sub("^[^\(]*\(([^\)]*)\).*$","\\1",rootcall)
            call_params=re.split(",",call_params)
            top_calls.append((call,call_params))
            # do implicit quantification
            process_input.add_implicit_exists(ex_params,globalfreevars,call_params)
        type=2
    else:
        # OLD version just for compatibility reasons
        # get parameters
        if not (re.search("^Params",contains[0])):
            raise process_input.InputError("No \"Params\" specified on 1st (nonempty) line of input")
        params=re.sub("^Params","",contains[0])
        params=re.split(",",params)
        del contains[0]
        # get root rule identifier
        if not (re.search("^Root",contains[0])):
            raise process_input.InputError("No \"Root\" specified on 2st (nonempty) line of input")
        root_rule=re.sub("^Root","",contains[0])
        del contains[0]
        type=0
    #Parse predicates
    empty_rule=0
    for x in contains:
        empty_rule=empty_rule+parse_predicate(x,preds)
    if empty_rule:
        # empty rules in the system of predicates -> inline them + create a formula for empty heap
        emptyheap_eq=process_input.inline_empty_rules(preds,top_calls)
    else:
        emptyheap_eq=[] # No empty heap defined by the system of predicates ---> false represented as []
    if type==2:
        # type==2: join operator on top level calls to translate into a single Rootcall
        (root_rule,params,emptyheap_eq)=join_core.join(preds,top_calls,emptyheap_eq,ex_params)
        # rename all variables in conflict between parameters and predicates
        process_input.rename_conflicts_with_params(preds,params)
        #track and eliminate all parameters
        for i in range(0,len(params)):
            ex_quantif=(params[i] in ex_params) # or (params[i]=="nil") # nil is allways handled as existentially quantified variable
            root_rule=process_input.trackeliminate(preds,root_rule,0,params[i],ex_quantif)
        # remove unreachable predicates
        process_input.remove_unreachable_predicates(preds,root_rule)
        # remove "nil" from params
        new_params=[]
        for i in params:
            if i!="nil" and not (i in ex_params):
                new_params.append(i)
        params=new_params
        # remove ex_params from emptyheap_eq
        new_emptyheap_eq=[]
        for disj in emptyheap_eq:
            new_disj=[]
            for conj in disj:
                new_conj=[]
                for x in conj:
                    if not x in ex_params:
                        new_conj.append(x)
                if len(new_conj)>1:
                    new_disj.append(new_conj)
            new_emptyheap_eq.append(new_disj)
        emptyheap_eq=new_emptyheap_eq
    else:
        # OLD version, just for compatibility reasons (type==0)
        if not emptyheap_eq==[]:
            emptyheap_eq=emptyheap_eq[0]
    sig=process_input.compute_signature(preds)
    aut,eq_edges=process_input.sl2ta(preds,sig,params,tiles,root_rule)
    return aut,emptyheap_eq,eq_edges
		
